// -*- C++ -*-
//
//  This file is part of CBNSIM, the Content-Based Networking
//  Simulation Library.
//
//  Authors: See the file AUTHORS for full details.
//
//  Copyright (C) 2002-2004 University of Colorado
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307,
//  USA, or send email to serl@cs.colorado.edu.
//
// $Id: CBNSimulation.cc,v 1.10 2005/10/17 16:11:05 hallcp Exp $
//
#include <cmath> // for rint()

#include <fstream>
#include <iostream>
#include "cbnsim/CBNSimulation.h"
#include "cbnsim/KillerProcess.h"
#include "cbnsim/io_exception.h"
#include "cbnsim/logging.h"

using namespace cbnsim;
using namespace std;

CBNSimulation :: CBNSimulation()
  : m_data_collector( NULL ), m_driver( NULL ), m_node_factory( NULL )
{
}

CBNSimulation :: ~CBNSimulation()
{
  delete m_data_collector;
  delete m_driver;
  delete m_node_factory;
}

void CBNSimulation :: debug_messages( bool debug_on )
{
  cbnsim::g_debug_on = debug_on;
}

bool CBNSimulation :: debug_messages() const
{
  return cbnsim::g_debug_on;
}

void CBNSimulation :: error_messages( bool error_on )
{
  cbnsim::g_error_on = error_on;
}

bool CBNSimulation :: error_messages() const
{
  return cbnsim::g_error_on;
}

void CBNSimulation :: info_messages( bool info_on )
{
  cbnsim::g_info_on = info_on;
}

bool CBNSimulation :: info_messages() const
{
  return cbnsim::g_info_on;
}

void CBNSimulation :: input_fname( const string& ifname )
{
  m_input_fname = ifname;
}

const string& CBNSimulation :: input_fname() const
{
  return m_input_fname;
}

void CBNSimulation :: output_fname( const string& ofname )
{
  m_output_fname = ofname;
}

const string& CBNSimulation :: output_fname() const
{
  return m_output_fname;
}

void CBNSimulation :: run() throw ( exception )
{
  CBNASSERT( m_node_factory,
	     "node_factory cannot be NULL" );
  /*
   * deal with the input file.
   */
  ifstream ifs;
  if( m_input_fname.size() == 0 )
    {
      m_parser.set_istream( cin );
    }
  else
    {
      ifs.open( m_input_fname.c_str() );
      if( ! ifs )
	{
	  throw io_exception( string( "unable to open: " ) + m_input_fname );
	}
      m_parser.set_istream( ifs );
    }
  /*
   * now deal with the output file
   */
  ostream* p_out = &cout;
  ofstream ofs;
  if( m_output_fname.size() )
    {
      ofs.open( m_output_fname.c_str() );
      if( ! ofs )
	{
	  throw io_exception( string( "unable to open: " ) + m_output_fname );
	}
      p_out = &ofs;
    }
  /*
   * parse the topology section.
   */
  m_parser.parse_topology();

  /*
   * set up the topology
   */
  setup_topology();

  /*
   * activate the killer.
   */
  KillerProcess killer( m_parser.get_sim_length() );
  killer.activate();

  /*
   * activate the driver.
   */
  if( ! m_driver )
    {
      m_driver = new Driver();
      CBNASSERT( m_driver,
		 "unable to allocate cbnsim::Driver" );
    }
  m_driver->setup( m_nodes, m_parser );
  m_driver->activate();

  /*
   * activate the data collector
   */
  if( ! m_data_collector )
    {
      m_data_collector = new DataCollector();
      CBNASSERT( m_data_collector,
		 "unable to allocate cbnsim::DataCollector" );
    }
  // I'm using rint instead of round because free fucking BSD doesn't
  // have it... -A
  //
  m_data_collector->setup( m_nodes, (ssim::Time)rint( m_dci / m_parser.get_time_unit() ), (*p_out) );
  m_data_collector->activate();

  pre_run_simulation_hook();
  /*
   * run the simulation.
   */
  ssim::Sim::run_simulation();

  /*
   * finalize the data collector
   */
  m_data_collector->finalize();
}

void CBNSimulation :: setup_link( unsigned int from,
				  unsigned int to,
				  unsigned int length )
{
  cout << "from: " << from << " to: " << to << " size: " << m_nodes.size() << endl;

  CBNASSERT( from < m_nodes.size(),
	     "from node id out of range: " << from );
  CBNASSERT( to < m_nodes.size(),
	     "to node id out of range: " << to );

  Node* fn = m_nodes[from];
  Node* tn = m_nodes[to];

  fn->add_neighbor( tn->pid(), length );
  tn->add_neighbor( fn->pid(), length );

  setup_link_hook( fn->pid(), tn->pid(), length );
  setup_link_hook( tn->pid(), fn->pid(), length );
}

void CBNSimulation :: setup_node(unsigned int i)
{
    Node* node = m_node_factory->newNode(i, m_parser.get_node_x(i),
					 m_parser.get_node_y(i));
    node->activate();
    m_nodes.push_back( node );
    setup_node_hook( node );
}

void CBNSimulation :: setup_topology()
{
  for( unsigned int i = 0; i < m_parser.get_node_count(); ++i )
    {
      setup_node(i);
    }
  
  set< pair<unsigned int,unsigned int> >::const_iterator i;
  for( i = m_parser.get_links().begin(); i != m_parser.get_links().end(); ++i )
    {
      unsigned int from = (*i).first;
      unsigned int to = (*i).second;
      setup_link( from, to, m_parser.get_link_weight( from, to ) );
    }
  setup_topology_hook();
}
