// -*- C++ -*-
//
//  This file is part of CBNSIM, the Content-Based Networking
//  Simulation Library.
//
//  Authors: See the file AUTHORS for full details.
//
//  Copyright (C) 2002-2004 University of Colorado
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307,
//  USA, or send email to serl@cs.colorado.edu.
//
// $Id: WorkloadGenerator.cc,v 1.5 2011-08-25 12:29:35 carzanig Exp $
//
#include <cstdio> // for fopen, etc.
#include <climits> // for PATH_MAX
#include <cstring>
#include <cstdlib>
#include <cmath>
#include <iostream>
#include <set>
#include <siena/ssbg.h>
#include <stdexcept>
#include <sys/time.h>
#include <unistd.h> // getcwd

#include <config.h>

#include "cbnsim/KillerProcess.h"
#include "cbnsim/WorkloadGenerator.h"

#define DEFAULT_DIST_DIR PREFIX "/share/" PACKAGE "-" VERSION "/dist"
#define DEFAULT_TOPO_DIR PREFIX "/share/" PACKAGE "-" VERSION "/topo"

using namespace cbnsim;
using namespace std;

extern FILE* yyin;
int yyparse( void );

extern map<int, set<int> > topo_adjacency_list;
extern map<pair<int,int>,double> topo_link_lengths;
extern map<int,pair<int,int> > topo_node_coords;

WorkloadGenerator :: WorkloadGenerator( ostream& out )
    : m_attr_max( 10 ),
      m_attr_min( 1 ),
      m_constr_max( 5 ),
      m_constr_min( 1 ),
      m_dist_dir( DEFAULT_DIST_DIR ),
      m_filt_max( 20 ),
      m_filt_min( 0 ),
      m_out( out ),
      m_p_cmd( "0" ),
      m_reuse_types( false ),
      m_s_cmd( "0" ),
      m_sim_length_min( 0 ),
      m_ttu_sec( 0.001 ),
      m_wtu_sec( 1.0 )
{
    struct timeval tv;
    gettimeofday(&tv, NULL);
    m_seed = tv.tv_usec;
}

WorkloadGenerator :: ~WorkloadGenerator()
{
    for( vector<WorkloadNodeProcessFactory*>::const_iterator i = m_npfs.begin();
	 i != m_npfs.end(); ++i )
    {
	delete (*i);
    }
}

void WorkloadGenerator :: add( WorkloadNodeProcessFactory* wnpf )
{
    m_npfs.push_back( wnpf );
}

int WorkloadGenerator :: main( const string& prog, const vector<string>& args ) throw ( exception )
{
    try
    {
	parse_args( args );
    }
    catch( argument_exception e )
    {
	cerr << "ERROR: INVALID ARGUMENT: " << e.what() << endl;
	print_usage();
	return 1;
    }
    /*
     * read in the topology info
     */
    uint topo_node_count = 0;
    {
	yyin = fopen( m_topo_file.c_str(), "r" );
	if( !yyin )
	{
	    string topo2( string( DEFAULT_TOPO_DIR ) + string( "/" ) + string( m_topo_file ) );
	    yyin = fopen( topo2.c_str(), "r" );
	    if( !yyin )
	    {
		cerr << "ERROR: unable to open topology file '" << m_topo_file << "'" << endl;
		cerr << "ERROR: unable to open topology file '" << topo2 << "'" << endl;
		return 1;
	    }
	    else
	    {
		m_topo_file = topo2;
	    }
	}
	do
	{
	    yyparse();
	}
	while( !feof( yyin ) );
	fclose( yyin );
	topo_node_count = topo_adjacency_list.size();
    }

    uint p = 0;		// the number of publisher nodes in the simulation.
    uint s = 0;		// the number of subscriber nodes in the simulation.

    if( m_p_cmd == "ALL" )
    {
	p = topo_node_count;
    }
    else if( (*m_p_cmd.rbegin()) == '%' )
    {
	float percent = atoi( m_p_cmd.substr( 0, m_p_cmd.size() - 1 ).c_str() ) / 100.0;
	// see comment in CBNSimulation.cc about rint() vs. round()
	p = (uint)rint( percent * topo_node_count );
    }
    else
    {
	p = atoi( m_p_cmd.c_str() );
	if( p > topo_node_count )
	{
	    cerr << "ERROR: number of publishers (" << p
		 << ") is larger than nodes in topo file ("
		 << topo_node_count << ")" << endl;
	    return 1;
	}
    }
      
    if( m_s_cmd == "ALL" )
    {
	s = topo_node_count;
    }
    else if( (*m_s_cmd.rbegin()) == '%' )
    {
	float percent = atoi( m_s_cmd.substr( 0, m_s_cmd.size() - 1 ).c_str() ) / 100.0;
	s = (uint)rint( percent * topo_node_count );
    }
    else
    {
	s = atoi( m_s_cmd.c_str() );
	if( s > topo_node_count )
	{
	    cerr << "ERROR: number of subscribers (" << s
		 << ") is larger than nodes in topo file ("
		 << topo_node_count << ")" << endl;
	    return 1;
	}
    }
    /*
     * random seed
     */
    srand( m_seed );

    /*
     * summarize command line arguments to standard out
     */
    char cwd[PATH_MAX];
    if (!getcwd( cwd, PATH_MAX )) {
	strncpy(cwd, "(unknown)", PATH_MAX);
    }

    m_out << "# " << prog << " invoked with the following settings:" << endl
	  << "# ssbg attr-min = " << m_attr_min << endl
	  << "# ssbg attr-max = " << m_attr_max << endl
	  << "# ssbg constr-min = " << m_constr_min << endl
	  << "# ssbg constr-max = " << m_constr_max << endl
	  << "# ssbg filt-min = " << m_filt_min << endl
	  << "# ssbg filt-max = " << m_filt_max << endl
	  << "# cwd = " << cwd << endl
	  << "# node count = " << topo_node_count << endl
	  << "# publisher command = " << m_p_cmd << endl
	  << "# publisher count = " << p << endl
	  << "# ssbg reuse-types = " << m_reuse_types << endl
	  << "# seed = " << m_seed << endl
	  << "# simulation length = " << m_sim_length_min << " (min)" << endl
	  << "# simulation time unit = " << m_wtu_sec << " (sec) " << endl
	  << "# ssbg dist dir = " << m_dist_dir << endl
	  << "# subscriber count = " <<  s << endl
	  << "# subscriber command = " << m_s_cmd << endl
	  << "# topology time unit = " << m_ttu_sec << " (sec) " << endl
	  << "# topology file = " << m_topo_file << endl
	  << "# node process settings:" << endl
	;
    /*
     * print out the settings for the node process factories
     */
    for( vector<WorkloadNodeProcessFactory*>::const_iterator j = m_npfs.begin();
	 j != m_npfs.end(); ++j )
    {
	(*j)->print_setting_summary( m_out );
    }

    /*
     * calculate the number of timesteps in the simulation
     */
    uint ts = m_sim_length_min * (uint)(60 / m_wtu_sec);

    /*
     * calculate the multiplier needed to bring topology timings into
     * the same time unit as the workload. This is used to scale the
     * link delays.
     */
    double factor = m_ttu_sec / m_wtu_sec;

    /*
     * set some output flags.
     */
    m_out << showpoint;
    m_out << boolalpha;

    /*
     * output the topology section
     */
    m_out << "time_unit " << m_wtu_sec << " ;" << endl;
    m_out << "sim_length " << ts << " ;" << endl;

    map<int,pair<int,int> >::const_iterator ci;
    for( ci = topo_node_coords.begin(); ci != topo_node_coords.end(); ++ci )
    {
	if( ci != topo_node_coords.begin() )
	{
	    m_out << " ," << endl;
	}
	m_out << "node " << (*ci).first << " "
	      << (*ci).second.first << " "
	      << (*ci).second.second;
    }
    cout << " ;" << endl;

    map<pair<int,int>,double>::const_iterator li;
    for( li = topo_link_lengths.begin(); li != topo_link_lengths.end(); ++li )
    {
	if( li != topo_link_lengths.begin() )
	{
	    m_out << " ," << endl;
	}
	m_out << "link " << (*li).first.first << " "
	      << (*li).first.second << " "
	      << (int)floor( factor * (*li).second );
    }
    m_out << " ;" << endl;

    /*
     * initialize the SSBG object.
     */
    ssbg::SSBG ssbg( m_attr_min, m_attr_max, m_constr_min, m_constr_max, 
		     m_filt_min, m_filt_max, m_reuse_types );

    ssbg.load_distributions( string( m_dist_dir + string( "/bool_values.dist" ) ).c_str(),
			     string( m_dist_dir + string( "/double_values.dist" ) ).c_str(),
			     string( m_dist_dir + string( "/int_values.dist" ) ).c_str(),
			     string( m_dist_dir + string( "/string_values.dist" ) ).c_str(),
			     string( m_dist_dir + string( "/types.dist" ) ).c_str(),
			     string( m_dist_dir + string( "/bool_operators.dist" ) ).c_str(),
			     string( m_dist_dir + string( "/double_operators.dist" ) ).c_str(),
			     string( m_dist_dir + string( "/int_operators.dist" ) ).c_str(),
			     string( m_dist_dir + string( "/string_operators.dist" ) ).c_str(),
			     string( m_dist_dir + string( "/attr_names.dist" ) ).c_str(),
			     string( m_dist_dir + string( "/constr_names.dist" ) ).c_str() );
      
    // need to pick the sets of publishers and 
    // subscribers.
    set<int> pset;
    set<int> sset;
    if( m_p_cmd == "ALL" )
    {
	for( uint i=0; i<p; ++i )
	{
	    pset.insert( i );
	}
    }
    else
    {
	while( pset.size() < p )
	{
	    pset.insert( ssbg::rand_normal( topo_node_count ) );
	}
    }
    if( m_s_cmd == "ALL" )
    {
	for( uint i=0; i<s; ++i )
	{
	    sset.insert( i );
	}
    }
    else
    {
	while( sset.size() < s )
	{
	    sset.insert( ssbg::rand_normal( topo_node_count ) );
	}
    }

    /*
     * Allocate all the processes 
     */
    vector<WorkloadNodeProcess*> node_processes;
    for( uint i = 0;  i < topo_node_count; ++i )
    {
	for( vector<WorkloadNodeProcessFactory*>::const_iterator j = m_npfs.begin();
	     j != m_npfs.end(); ++j )
	{
	    bool needs_process = false;
	    if( (*j)->applies_to( WorkloadNodeProcessFactory::Any_Node ) )
	    {
		needs_process = true;
	    }
	    if( !needs_process && pset.find( i ) != pset.end() && 
		(*j)->applies_to( WorkloadNodeProcessFactory::Publisher_Node ) )
	    {
		needs_process = true;
	    }
	    if( !needs_process && sset.find( i ) != sset.end() &&
		(*j)->applies_to( WorkloadNodeProcessFactory::Subscriber_Node ) )
	    {
		needs_process = true;
	    }

	    if( needs_process )
	    {
		WorkloadNodeProcess *np = (*j)->newWorkloadNodeProcess( i, m_out, ssbg, m_wtu_sec );
		np->activate();
		node_processes.push_back( np );
	    }
	}
    }

    /*
     * create and start the killer process that will terminate the
     * simulation.
     */
    KillerProcess killer( ts );
    killer.activate();
  
    ssim::Sim::run_simulation();
    return 0;
}

void WorkloadGenerator :: parse_args( const vector<string>& args )
    throw ( argument_exception )
{
    for( vector<string>::const_iterator i = args.begin(); i != args.end(); ++i )
    {
	if( (*i) == "-p" )
	{
	    if( ++i != args.end() )
	    {
		m_p_cmd = (*i);
	    }
	    else
	    {
		throw argument_exception( "-p requires an argument" );
	    }
	}
	else if( (*i) == "-s" )
	{
	    if( ++i != args.end() )
	    {
		m_s_cmd = (*i);
	    }
	    else
	    {
		throw argument_exception( "-s requires an argument" );
	    }
	}
	else if( (*i) == "-seed" )
	{
	    if( ++i != args.end() )
	    {
		m_seed = atol( (*i).c_str() );
	    }
	    else
	    {
		throw argument_exception( "-seed requires an argument" );
	    }
	}
	else if( (*i) == "-l" )
	{
	    if( ++i != args.end() )
	    {
		m_sim_length_min = atoi( (*i).c_str() );
	    }
	    else
	    {
		throw argument_exception( "-l requires an argument" );
	    }
	}
	else if( (*i) == "-wtu" )
	{
	    if( ++i != args.end() )
	    {
		m_wtu_sec = atof( (*i).c_str() );
	    }
	    else
	    {
		throw argument_exception( "-wtu requires an argument" );
	    }
	}
	else if( (*i) == "-topo" )
	{
	    if( ++i != args.end() )
	    {
		m_topo_file = (*i);
	    }
	    else
	    {
		throw argument_exception( "-topo requires an argument" );
	    }
	}
	else if( (*i) == "-ttu" )
	{
	    if( ++i != args.end() )
	    {
		m_ttu_sec = atof( (*i).c_str() );
	    }
	    else
	    {
		throw argument_exception( "-ttu requires an argument" );
	    }
	}
	else if( (*i) == "-dist-dir" )
	{
	    if( ++i != args.end() )
	    {
		m_dist_dir = (*i).c_str();
	    }
	    else
	    {
		throw argument_exception( "-dist-dir requires an argument" );
	    }
	}
	else if( (*i) ==  "-filt-min" )
	{
	    if( ++i != args.end() )
	    {
		m_filt_min = atoi( (*i).c_str() );
	    }
	    else
	    {
		throw argument_exception( "-filt-min requires an argument" );
	    }
	}
	else if( (*i) ==  "-filt-max" )
	{
	    if( ++i != args.end() )
	    {
		m_filt_max = atoi( (*i).c_str() );
	    }
	    else
	    {
		throw argument_exception( "-filt-max requires an argument" );
	    }
	}
	else if( (*i) ==  "-attr-min" )
	{
	    if( ++i != args.end() )
	    {
		m_attr_min = atoi( (*i).c_str() );
	    }
	    else
	    {
		throw argument_exception( "-attr-min requires an argument" );
	    }
	}
	else if( (*i) ==  "-attr-max" )
	{
	    if( ++i != args.end() )
	    {
		m_attr_max = atoi( (*i).c_str() );
	    }
	    else
	    {
		throw argument_exception( "-attr-max requires an argument" );
	    }
	}
	else if( (*i) ==  "-constr-min" )
	{
	    if( ++i != args.end() )
	    {
		m_constr_min = atoi( (*i).c_str() );
	    }
	    else
	    {
		throw argument_exception( "-constr-min requires an argument" );
	    }
	}
	else if( (*i) ==  "-constr-max" )
	{
	    if( ++i != args.end() )
	    {
		m_constr_max = atoi( (*i).c_str() );
	    }
	    else
	    {
		throw argument_exception( "-constr-max requires an argument" );
	    }
	}
	else
	{
	    // unhandled argument, pass off to process factories
	    bool handled = false;
	    for( vector<WorkloadNodeProcessFactory*>::const_iterator j = m_npfs.begin();
		 j != m_npfs.end(); ++j )
	    {
		if( (*j)->handle_argument( i, args.end() ) )
		{
		    handled = true;
		}
	    }

	    if( !handled )
	    {
		throw argument_exception( (*i).c_str() );
	    }
	}
    }
    /*
     * check required arguments
     */
    if( m_p_cmd.empty() )
    {
	throw argument_exception( "'-p' is a required parameter" );
    }
    if( m_s_cmd.empty() )
    {
	throw argument_exception( "'-s' is a required parameter" );
    }
    if( !m_sim_length_min )
    {
	throw argument_exception( "'-l' is a required parameter" );
    }
    if( m_wtu_sec > 1.0 )
    {
	// the calculation of ts_per_min will screw up if 60 / wtu does not
	// result in an integer, so at this point its just easiest
	// to force it to be <= 1.0
	throw argument_exception( "'-wtu' cannot be greater than 1.0" );
    }
    if( m_topo_file.size() == 0 )
    {
	throw argument_exception( "'-topo' is a required parameter" );
    }

    for( vector<WorkloadNodeProcessFactory*>::const_iterator j = m_npfs.begin();
	 j != m_npfs.end(); ++j )
    {
	if( ! (*j)->satisfied() )
	{
	    throw argument_exception( (*j)->event_name() + string( " missing required argument(s)" ) );
	}
    }
}

void WorkloadGenerator :: print_usage( const char* e )
{
    if( e != NULL )
    {
	m_out << "ERROR: " << e << endl;
    }
    m_out << "usage: [options...]\n\
required parameters:\n\
\t -p <int>|ALL|<real>%  number of publisher nodes in the simulation\n\
\t -s <int>|ALL|<real>%  number of subscriber nodes in the simulation\n\
\t -l <min>              length of the simulation in minutes\n\
\t -topo <file>          the file name of the topology file\n\
\t                       if the specified file cannot be found\n\
\t                       then directory '" << DEFAULT_TOPO_DIR << "' is searched\n\n";

    /*
     * print out the settings for the node process factories
     */
    for( vector<WorkloadNodeProcessFactory*>::const_iterator j = m_npfs.begin();
	 j != m_npfs.end(); ++j )
    {
	(*j)->print_required_usage( m_out );
    }

    m_out << 
	"\noptional parameters:\n\
\t -wtu <sec>            the workload time unit (s) (defaults to 1.0)\n\
\t -ttu <sec>		 the topology latency time unit (s) (defaults to 0.001)\n\
\t -dist-dir <dir>       the directory that contains the necessary SSBG dist files\n\
\t                       defaults to '" << DEFAULT_DIST_DIR << "'\n\
\t -seed <num>           random seed (defaults to current time)\n\
\t -attr-min <num>	 sets the minimum number of attributes per message (defaults to 1)\n\
\t -attr-max <num>	 sets the maximum number of attributes per message (defaults to 10)\n\
\t -constr-min <num>	 sets the minimum number of constraints per predicate (defaults to 1)\n\
\t -constr-max <num>	 sets the maximum number of constraints per predicate (defaults to 5)\n\
\t -filt-min <num>	 sets the minimum number of filters per predicate (defaults to 0)\n\
\t -filt-max <num>	 sets the maximum number of filters per predicate (defaults to 20)\n\
" << endl;
    /*
     * print out the settings for the node process factories
     */
    for( vector<WorkloadNodeProcessFactory*>::const_iterator j = m_npfs.begin();
	 j != m_npfs.end(); ++j )
    {
	(*j)->print_optional_usage( m_out );
    }

}

